#!/usr/bin/ruby

# collect_diffs.rb expects to find this script in the same directory as it
#

# TODO: generate subjectPrefix from repository dir
# TODO: exemplify syntax for 'cvs admin -m' when log message is missing

def blah(text)
  if $debug
    puts("cvsspam.rb: #{text}")
  end
end

class LogReader
  def initialize(logIO)
    @io = logIO
    advance
  end

  def currentLineCode ; @line[1,1]  end


  class ConstrainedIO
    def initialize(reader)
      @reader = reader
      @linecode = reader.currentLineCode
    end

    def each
      return if @reader == nil
      while true
        yield @reader.currentLine
        break unless @reader.advance && currentValid?
      end
      @reader = nil
    end

    def gets
      return nil if @reader == nil
      line = @reader.currentLine
      return nil if line==nil || !currentValid?
      @reader.advance
      return line
    end

    def currentValid?
      @linecode == @reader.currentLineCode
    end
  end

  def getLines
    ConstrainedIO.new(self)
  end

  def eof ; @line==nil  end

  def advance
    @line = @io.gets
    return false if @line == nil
    unless @line[0,1] == "#"
      raise "#{$logfile}:#{@io.lineno} line did not begin with '#': #{@line}"
    end
    return true
  end

  def currentLine
    @line==nil ? nil : @line[3, @line.length-4]
  end
end


def htmlEncode(text)
  text.gsub(/./) do
    case $&
      when "&" then "&amp;"
      when "<" then "&lt;"
      when ">" then "&gt;"
      when "\"" then "&quot;"
      else $&
    end
  end
end

# actually, allows '/' to appear
def urlEncode(text)
  text.sub(/[^a-zA-Z0-9\-,.*_\/]/) do
    "%#{sprintf('%2X', $&[0])}"
  end
end

class FileEntry
  def initialize(path)
    @path = path
    @lineAdditions = @lineRemovals = 0
  end

  attr_accessor :path, :type, :lineAdditions, :lineRemovals, :isBinary, :isEmpty

  def file
    @path =~ /(.*)\/(.*)/
    $2
  end

  def basedir
    @path =~ /(.*)\/(.*)/
    $1
  end

  def removal?
    @type == "R"
  end

  def addition?
    @type == "A"
  end

  def modification?
    @type == "M"
  end
end


class LineConsumer
  def handleLines(lines, emailIO)
    @emailIO = emailIO
    @lineCount = 0
    setup
    lines.each do |line|
      @lineCount += 1
      consume(line)
    end
    teardown
  end

  def setup
  end

  def teardown
  end

  def lineno
    @lineCount
  end

  def println(text)
    @emailIO.puts(text)
  end

  def print(text)
    @emailIO.print(text)
  end
end


class CommentHandler < LineConsumer
  def setup
    @haveBlank = false
    @comment = ""
  end

  def consume(line)
    if line =~ /^\s*$/
      @haveBlank = true
    else
      if @haveBlank
        @comment += "\n"
        @haveBlank = false
      end
      $mailSubject = line if $mailSubject == nil
      @comment += line += "\n"
    end
  end

  def teardown
    unless @comment == @lastComment
      println("<pre class=\"comment\">")
      encoded = htmlEncode(@comment)
      unless $bugzillaURL == nil
        # make links to bugzilla bugs -- text like 'bug 1234'
        encoded.gsub!(/\b(bug\s+([0-9]+))/) do
	  text = $1
	  "<a href=\"#{$bugzillaURL.sub(/%s/, $2)}\">#{text}</a>"
	end
      end
      println(encoded)
      println("</pre>")
      @lastComment = @comment
    end
  end
end


class TagHandler < LineConsumer
  def consume(line)
    # TODO: check there is only one line
    @tag = line
  end

  def getLastTag
    tmp = @tag
    @tag = nil
    tmp
  end
end

class VersionHandler < LineConsumer
  def consume(line)
    # TODO: check there is only one line
    $fromVer,$toVer = line.split(/,/)
  end
end


class FileHandler < LineConsumer
  def setTagHandler(handler)
    @tagHandler = handler
  end

  def setup
    println("<hr noshade size=\"1\" /><a name=\"file#{$fileEntries.size+1}\" /><div class=\"file\">")
  end

  def consume(line)
    $file = FileEntry.new(line)
    $fileEntries << $file
    findCommonPath($file.basedir)
    handleFile($file)
  end

 protected
  def getTag
    @tagHandler.getLastTag
  end

  def findCommonPath(path)
    path = path.dup
    if $commonPath == nil
      $commonPath = path
    else
      until $commonPath == path
        if $commonPath.size>path.size
	  $commonPath.sub!(/(.*)\/.*$/, '\1')
        else
	  path.       sub!(/(.*)\/.*$/, '\1')
        end
      end
    end
  end

  # if using viewcvs, create a link to the named path's dir listing
  def viewcvsPath(path)
    if $viewcvsURL == nil
      htmlEncode(path)
    else
      "<a href=\"#{$viewcvsURL}#{urlEncode(path)}\">#{htmlEncode(path)}</a>"
    end
  end

  def viewcvsFile(path, version)
    if $viewcvsURL == nil
      version
    else
      "<a href=\"#{$viewcvsURL}#{urlEncode(path)}?rev=#{version}&content-type=text/vnd.viewcvs-markup\">#{version}</a>"
    end
  end
end


# in need of refactoring...

class AddedFileHandler < FileHandler
  def handleFile(file)
    file.type="A"
    println("<div class=\"pathname\" id=\"added\">")
    println(viewcvsPath(file.basedir))
    println("</div>")
    println("<div class=\"filename\" id=\"added\">#{htmlEncode(file.file)} <span id=\"info\">added at #{viewcvsFile(file.path,$toVer)}</span></div>")
  end
end

class RemovedFileHandler < FileHandler
  def handleFile(file)
    file.type="R"
    println("<div class=\"pathname\" id=\"removed\">")
    println(viewcvsPath(file.basedir))
    println("</div>")
    println("<div class=\"filename\" id=\"removed\">#{htmlEncode(file.file)} <span id=\"info\">removed after #{viewcvsFile(file.path,$fromVer)}</span></div>")
  end
end

class ModifiedFileHandler < FileHandler
  def handleFile(file)
    file.type="M"
    println("<div class=\"pathname\">")
    println(viewcvsPath(file.basedir))
    println("</div>")
    println("<div class=\"filename\">#{htmlEncode(file.file)} <span id=\"info\">#{viewcvsFile(file.path,$fromVer)} #{viewcvsDiff(file.path,$fromVer,$toVer)} #{viewcvsFile(file.path,$toVer)}</span></div>")
  end

  def viewcvsDiff(path,from,to)
    if $viewcvsURL == nil
      "->"
    else
      "<a href=\"#{$viewcvsURL}#{urlEncode(path)}.diff?r1=#{from}&r2=#{to}\">-></a>"
    end
  end
end


class UnifiedDiffStats
  def consume(line)
    case line[0,1]
      when "+" then $file.lineAdditions += 1
      when "-" then $file.lineRemovals += 1
    end
  end
end

# TODO: change-within-line colourisation should really be comparing the
#       set of lines just removed with the set of lines just added, but
#       it currently considers just a single line

class UnifiedDiffColouriser < LineConsumer
  def output=(io)
    @emailIO = io
  end

  def consume(line)
    initial = line[0,1]
    if initial != @currentState
      prefixLen = 1
      suffixLen = 0
      if initial=="+" && @currentState=="-"
        # may be an edit, try to highlight the changes part of the line
        a = line[1,line.length-1]
        b = @lineJustDeleted[1,@lineJustDeleted.length-1]
        prefixLen = commonPrefixLength(a, b)+1
        suffixLen = commonPrefixLength(a.reverse, b.reverse)
        if prefixLen==1 && suffixLen==0 || prefixLen+suffixLen==@lineJustDeleted.length
          println(htmlEncode(@lineJustDeleted))
        else
          print(htmlEncode(@lineJustDeleted[0,prefixLen]))
          print("<span id=\"removedchars\">")
puts "@lineJustDeleted.length=#{@lineJustDeleted.length} prefixLen=#{prefixLen} suffixLen=#{suffixLen}"
          print(htmlEncode(@lineJustDeleted[prefixLen,@lineJustDeleted.length-prefixLen-suffixLen]))
          print("</span>")
          println(htmlEncode(@lineJustDeleted[@lineJustDeleted.length-suffixLen,suffixLen]))
        end
      end
      if initial=="-"
        @lineJustDeleted=line
	shift(initial)
        # we'll print it next time (fingers crossed)
        return
      else
        @lineJustDeleted = nil
      end
      shift(initial)
      if prefixLen==1 && suffixLen==0 || prefixLen+suffixLen==line.length
        encoded = htmlEncode(line)
      else
        encoded = htmlEncode(line[0,prefixLen]) +
        "<span id=\"addedchars\">" +
        htmlEncode(line[prefixLen,line.length-prefixLen-suffixLen]) +
        "</span>" +
        htmlEncode(line[line.length-suffixLen,suffixLen])
      end
    else
      encoded = htmlEncode(line)
    end
    if initial=="-"
      unless @lineJustDeleted==nil
        println(htmlEncode(@lineJustDeleted))
      end
      @lineJustDeleted=line
      # we'll print it next time (fingers crossed)
      return
    end
    if initial=="+"
      if line =~ /\b(TODO\b.*)/
        $todoList << $1
        encoded.sub!(/\bTODO\b/, "<span id=\"TODO\">TODO</span>")
        encoded = "<a name=\"todo#{$todoList.size}\" />#{encoded}"
      elsif line =~ /\b(FIXME\b.*)/
        $todoList << $1
        encoded.sub!(/\FIXME\b/, "<span id=\"TODO\">FIXME</span>")
        encoded = "<a name=\"todo#{$todoList.size}\" />#{encoded}"
      end
      # won't bother with /XXX/ (false positives more likely?)
    end
    println(encoded)
  end

  def teardown
    unless @lineJustDeleted==nil
      println(htmlEncode(@lineJustDeleted))
      @lineJustDeleted = nil
    end
  end

 private
  def shift(nextState)
    unless @currentState == nil
      println("</pre>")
      style = case nextState
        when "\\" then "info" # as in '\ No newline at end of file'
        when "@" then "info"
        when " " then "context"
        when "+" then "added"
        when "-" then "removed"
      end
      println("<pre class=\"diff\" id=\"#{style}\">")
    end
    @currentState = nextState
  end

  def commonPrefixLength(a, b)
    length = 0
    a.each_byte do |char|
      break unless b[length]==char
      length = length + 1
    end
    return length
  end
end


class UnifiedDiffHandler < LineConsumer
  def setup
    @stats = UnifiedDiffStats.new
    @colour = UnifiedDiffColouriser.new
    @colour.output = @emailIO
    @lookahead = nil
    # evil here:
    def self.consume(line)
      case lineno
        when 1
         @diffline = line
        when 2
          @lookahead = line
        when 3
	  # TODO: move to UnifiedDiffColouriser
	  println("<pre class=\"diff\" id=\"info\">")
	  println(htmlEncode(@diffline))  # 'diff ...'
	  println(htmlEncode(@lookahead)) # '--- ...'
	  println(htmlEncode(line))      # '+++ ...'

	  # now, redefine this method for the subsequent lines of input
	  def self.consume(line)
	    @stats.consume(line)
	    @colour.consume(line)
	  end
      end
    end
  end

  def teardown
    if @lookahead == nil
      println("[Empty file]")
      $file.isEmpty = true
    elsif @lookahead  =~ /Binary files .* and .* differ/
      println("[Binary file]")
      $file.isBinary = true
    else
      @colour.teardown
      println("</pre>")
    end
    println("</div>") # end of "file" div
  end
end









$config = "#{ENV['CVSROOT']}/CVSROOT/cvsspam.config"

$debug = false
$recipients = Array.new
$sendmail_prog = "/usr/sbin/sendmail"

require 'getoptlong'

opts = GetoptLong.new(
  [ "--to",     "-t", GetoptLong::REQUIRED_ARGUMENT ],
  [ "--config", "-c", GetoptLong::REQUIRED_ARGUMENT ],
  [ "--debug",  "-d", GetoptLong::NO_ARGUMENT ]
)

opts.each do |opt, arg|
  $recipients << arg if opt=="--to"
  $config = arg if opt=="--config"
  $debug = true if opt=="--debug"
end


if ARGV.length != 1
  if ARGV.length > 1
    $stderr.puts "extra arguments not needed: #{ARGV[1, ARGV.length-1].join(', ')}"
  else
    $stderr.puts "missing required file argument"
  end
  puts "Usage: cvsspam.rb [ --to <email> ] [ --config <file> ] <collect_diffs file>"
  exit(-1)
end

$logfile = ARGV[0]


$additionalHeaders = Array.new
$problemHeaders = Array.new

# helper functions called from the 'config file'
def addHeader(name, value)
  if name =~ /^[!-9;-~]+$/
    $additionalHeaders << [name, value]
  else
    $problemHeaders << [name, value]
  end
end
def addRecipient(email)
  $recipients << email
end

load $config

if $recipients.empty?
  fail "No email recipients defined"
end

$viewcvsURL << "/" if $viewcvsURL!=nil && $viewcvsURL!~/\/$/


tagHandler = TagHandler.new

$handlers = Hash[">" => CommentHandler.new,
		 "U" => UnifiedDiffHandler.new,
		 "T" => tagHandler,
		 "A" => AddedFileHandler.new,
		 "R" => RemovedFileHandler.new,
		 "M" => ModifiedFileHandler.new,
		 "V" => VersionHandler.new]

$handlers["A"].setTagHandler(tagHandler)
$handlers["R"].setTagHandler(tagHandler)
$handlers["M"].setTagHandler(tagHandler)

$fileEntries = Array.new
$todoList = Array.new

File.open("#{$logfile}.emailtmp", File::RDWR|File::CREAT|File::TRUNC) do |mail|

  File.open($logfile) do |log|
    reader = LogReader.new(log)

    until reader.eof
      handler = $handlers[reader.currentLineCode]
      if handler == nil
        raise "No handler file lines marked '##{reader.currentLineCode}'"
      end
      handler.handleLines(reader.getLines, mail)
    end
  end
end

$subjectPrefix = "[CVS #{#subjectTag}]"

puts "invoking #{$sendmail_prog} -t"
IO.popen("#{$sendmail_prog} -t", "w") do |mail|
  mail.puts("To: #{$recipients.join(',')}")
  mail.puts("Subject: #{$subjectPrefix} #{$mailSubject}")
  mail.puts("Content-Type: text/html")
  unless ($additionalHeaders.empty?)
    $additionalHeaders.each do |header|
      mail.puts("#{header[0]}: #{header[1]}")
    end
  end
  mail.puts  # end-of-headers

  mail.puts(<<HEAD)
<html>
<head>
<style>
  .file {border-width:1px;border-style:solid;border-color:#eeeeee;margin-top:1em;margin-bottom:1em;}
  .pathname {font-family:monospace;}
  .filename {font-family:monospace;font-weight:bolder;font-size:x-large;margin-bottom:.5em;}
  .diff {margin:0;}
  .todolist {padding:4px;border:1px;border-color:#000000;border-style:dashed;margin-top:1em;}
  .todolist ul {margin-top:0;margin-bottom:0;}
  #added {background-color:#ddffdd;}
  #addedchars {background-color:#88ff88;font-weight:bolder;}
  #removed {background-color:#ffdddd;}
  #removedchars {background-color:#ff8888;font-weight:bolder;}
  #info {color:#888888;font-size:small;}
  #context {background-color:#eeeeee;}
  #todo {background-color:#ffff00;}
  .comment {padding:4px;border:1px;border-color:#000000;border-style:dashed;background-color:#ffffdd}
  .error {color:red;}
</style>
</head>
<body>
HEAD

  unless ($problemHeaders.empty?)
    mail.puts("<strong class=\"error\">Bad header format in '#{ARGV[0]}':<ul>")
    $stderr.puts("Bad header format in '#{ARGV[0]}':")
    $problemHeaders.each do |header|
      mail.puts("<li><pre>#{htmlEncode(header[0])}</pre></li>")
      $stderr.puts(" - #{header[0]}")
    end
    mail.puts("</ul></strong>")
  end
  mail.puts("Commit in <b><tt>#{htmlEncode($commonPath)}</tt></b>")
  mail.puts("<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" rules=\"rows\" frame=\"void\">")
  filesAdded = 0
  filesRemoved = 0
  filesModified  = 0
  totalLinesAdded = 0
  totalLinesRemoved = 0
  count = 0
  $fileEntries.each do |file|
    count += 1
    if file.addition?
      filesAdded += 1
      mail.puts("<tr id=\"added\"><td>A&nbsp;</td>")
    elsif file.removal?
      filesRemoved += 1
      mail.puts("<tr id=\"removed\"><td>R&nbsp;</td>")
    elsif file.modification?
      filesModified += 1
      mail.puts("<tr><td>&nbsp;</td>")
    end
    name = file.path.slice($commonPath.size+1,file.path.size-$commonPath.size-1)
    mail.puts("<td><tt><a href=\"#file#{count}\">#{htmlEncode(name)}</a>&nbsp;</tt></td>")
    if file.isEmpty
      # this is possible if an empty file is added to CVS
      mail.puts("<td colspan=\"2\" align=\"center\">[empty]</td>")
    elsif file.isBinary
      mail.puts("<td colspan=\"2\" align=\"center\">[binary]</td>")
    else
      if file.lineAdditions>0
        totalLinesAdded += file.lineAdditions
        mail.print("<td align=\"right\" id=\"added\">+#{file.lineAdditions}</td>")
      else
	mail.print("<td>&nbsp;</td>")
      end
      if file.lineRemovals>0
        totalLinesRemoved += file.lineRemovals
        mail.print("<td align=\"right\" id=\"removed\">-#{file.lineRemovals}</td>")
      else
        mail.print("<td>&nbsp;</td>")
      end
      mail.puts("</td></tr>")
    end
  end
  if $fileEntries.size>1
    # give total number of lines added/removed accross all files
    mail.puts("<tr><td>&nbsp;</td><td>&nbsp;</td>")
    if totalLinesAdded>0
      mail.print("<td align=\"right\" id=\"added\">+#{totalLinesAdded}</td>")
    else
      mail.print("<td>&nbsp;</td>")
    end
    if totalLinesRemoved>0
      mail.print("<td align=\"right\" id=\"removed\">+#{totalLinesRemoved}</td>")
    else
      mail.print("<td>&nbsp;</td>")
    end
    mail.puts("</tr>")
  end
  
  mail.puts("</table>")

  if (filesAdded+filesRemoved+filesModified)>1
    mail.puts("<span id=\"info\">")
    if filesAdded>0
      mail.print("#{filesAdded} added")
    end
    if filesRemoved>0
      mail.print(" + ") if filesAdded>0
      mail.print("#{filesRemoved} removed")
    end
    if filesModified>0
      mail.print(" + ") if filesAdded>0 || filesRemoved>0
      mail.print("#{filesModified} modified")
    end
    mail.puts(" files</span><br>")
  end

  count = 0
  $todoList.each do |item|
    if count == 0
      mail.puts("<div class=\"todolist\"><center><span id=\"info\">Possibly new TODO items</span></center>")
      mail.puts("<ul>")
    end
    count += 1
    item = htmlEncode(item)
    mail.puts("<li><a href=\"#todo#{count}\">#{item}</a></li>")
  end
  mail.puts("</ul></div>") if count>0


  File.open("#{$logfile}.emailtmp") do |input|
    input.each do |line|
      mail.puts(line)
    end
  end
  if $debug
    blah("leaving file #{$logfile}.emailtmp")
  else
    File.unlink("#{$logfile}.emailtmp")
  end

  mail.puts("</body></html>")

end
